#!/usr/bin/env python

"""
Nonconformity functions.
"""

# Authors: Henrik Linusson

from __future__ import division

import numpy as np

# -----------------------------------------------------------------------------
# Classification error functions
# -----------------------------------------------------------------------------
def inverse_probability(prediction, y):
	"""Calculates the probability of not predicting the correct class.

	For each correct output in ``y``, nonconformity is defined as

	.. math::
		1 - \hat{P}(y_i | x) \, .

	Parameters
	----------
	prediction : numpy array of shape [n_samples, n_classes]
		Class probability estimates for each sample.

	y : numpy array of shape [n_samples]
		True output labels of each sample.

	Returns
	-------
	nc : numpy array of shape [n_samples]
		Nonconformity scores of the samples.
	"""
	prob = np.zeros(y.size, dtype=np.float32)
	for i, y_ in enumerate(y):
		prob[i] = prediction[i, y[i]]
	return 1 - prob

def margin(prediction, y):
	"""Calculates the margin error.

	For each correct output in ``y``, nonconformity is defined as

	.. math::
		0.5 - \hat{P}(y_i | x) - max_{y \, != \, y_i} \hat{P}(y | x) \, .

	Parameters
	----------
	prediction : numpy array of shape [n_samples, n_classes]
		Class probability estimates for each sample.

	y : numpy array of shape [n_samples]
		True output labels of each sample.

	Returns
	-------
	nc : numpy array of shape [n_samples, n_classes]
		Nonconformity scores for each sample and each class.
	"""
	prob = np.zeros(y.size, dtype=np.float32)
	for i, y_ in enumerate(y):
		prob[i] = prediction[i, y[i]]
		prediction[i, y[i]] = -np.inf
	return 0.5 - ((prob - prediction.max(axis=1)) / 2)

# -----------------------------------------------------------------------------
# Regression error functions
# -----------------------------------------------------------------------------
def abs_error(prediction, y):
	"""Calculates absolute error nonconformity for regression problems.

	For each correct output in ``y``, nonconformity is defined as

	.. math::
		| y_i - \hat{y}_i |

	Parameters
	----------
	prediction : numpy array of shape [n_samples]
		Regression prediction for each sample.

	y : numpy array of shape [n_samples]
		True output labels of each sample.

	Returns
	-------
	nc : numpy array of shape [n_samples]
		Nonconformity score of each sample.
	"""
	return np.abs(prediction - y)

def abs_error_inv(prediction, nc, significance):
	"""Calculates a prediction from an absolute-error nonconformity function.

	Calculates the partial inverse of the ``absolute_error`` nonconformity
	function, i.e., the minimum and maximum boundaries of the prediction
	interval for each point prediction, given a significance level.

	Parameters
	----------
	prediction : numpy array of shape [n_samples]
		Point (regression) predictions of a test examples.

	Returns
	-------
	interval : numpy array of shape [n_samples, 2]
		Minimum and maximum interval boundaries for each prediction.
	"""
	nc = np.sort(nc)[::-1]
	border = int(np.floor(significance * (nc.size + 1))) - 1
	# TODO: should probably warn against too few calibration examples
	border = max(border, 0)
	return np.vstack([prediction - nc[border], prediction + nc[border]]).T

def sign_error(prediction, y):
	"""Calculates signed error nonconformity for regression problems.

	For each correct output in ``y``, nonconformity is defined as

	.. math::
		y_i - \hat{y}_i

	Parameters
	----------
	prediction : numpy array of shape [n_samples]
		Regression prediction for each sample.

	y : numpy array of shape [n_samples]
		True output labels of each sample.

	Returns
	-------
	nc : numpy array of shape [n_samples]
		Nonconformity score of each sample.
	"""
	return prediction - y

def sign_error_inv(prediction, nc, significance):
	"""Calculates a prediction from a signed-error nonconformity function.

	Calculates the partial inverse of the ``signed_error`` nonconformity
	function, i.e., the minimum and maximum boundaries of the prediction
	interval for each point prediction, given a significance level.

	Parameters
	----------
	prediction : numpy array of shape [n_samples]
		Point (regression) predictions of a test examples.

	Returns
	-------
	interval : numpy array of shape [n_samples, 2]
		Minimum and maximum interval boundaries for each prediction.
	"""
	nc = np.sort(nc)[::-1]
	upper = int(np.floor((significance / 2) * (nc.size + 1)))
	lower = int(np.floor((1 - significance / 2) * (nc.size + 1)))
	# TODO: should probably warn against too few calibration examples
	upper = max(upper, 0)
	lower = min(lower, nc.size - 1)
	return np.vstack([prediction + nc[lower], prediction + nc[upper]]).T

# -----------------------------------------------------------------------------
# Base nonconformity scorer
# -----------------------------------------------------------------------------
class BaseNc(object):
	"""Base class for nonconformity scorers based on an underlying model.
	"""
	def __init__(self, model_class, err_func, model_params=None):
		self.err_func = err_func
		self.model_class = model_class
		self.model_params = model_params if model_params else {}
		self.model = self.model_class(**self.model_params)

		self.last_x, self.last_y = None, None
		self.last_prediction = None
		self.clean = False

	def fit(self, x, y):
		"""Fits the underlying model of the nonconformity scorer.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for fitting the underlying model.

		y : numpy array of shape [n_samples]
			Outputs of examples for fitting the underlying model.

		Returns
		-------
		None
		"""
		self.model.fit(x, y)
		self.clean = False

	def __get_prediction(self, x):
		if (not self.clean or
			self.last_x is None or
		    not np.array_equal(self.last_x, x)):

			self.last_x = x
			self.last_prediction = self._underlying_predict(x)
			self.clean = True

		return self.last_prediction.copy()

	def calc_nc(self, x, y):
		"""Calculates the nonconformity score of a set of samples.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for which to calculate a nonconformity score.

		y : numpy array of shape [n_samples, n_features]
			Outputs of examples for which to calculate a nonconformity score.

		Returns
		-------
		nc : numpy array of shape [n_samples]
			Nonconformity scores of samples.
		"""
		prediction = self.__get_prediction(x)
		return self.err_func(prediction, y)

# -----------------------------------------------------------------------------
# Classification nonconformity scorers
# -----------------------------------------------------------------------------
class ProbEstClassifierNc(BaseNc):
	"""Nonconformity scorer using an underlying class probability estimating
	model.

	Parameters
	----------
	model_class : class
		The model_class should be implement the ``fit(x, y)`` and
		``predict_proba(x)`` methods, as used by the classification models
		present in the scikit-learn library.

	err_func : callable
		Scorer callable object with signature ``score(estimator, x, y)``.

	model_params : dict, optional
		Dict containing keyword parameters to pass to model_class upon
		initialization.

	Attributes
	----------
	model_class : class
		Class used to construct the underlying model.

	err_func : callable
		Scorer function used to calculate nonconformity scores.

	model_params : dict
		Parameters sent to underlying model.

	model : object
		Underlying model object.

	See also
	--------
	RegressorNc
	"""
	def __init__(self,
	             model_class,
	             err_func,
	             model_params=None):
		super(ProbEstClassifierNc, self).__init__(model_class,
		                                          err_func,
		                                          model_params)

	def _underlying_predict(self, x):
		return self.model.predict_proba(x)

# -----------------------------------------------------------------------------
# Regression nonconformity scorers
# -----------------------------------------------------------------------------
class RegressorNc(BaseNc):
	"""Nonconformity scorer using an underlying regression model.

	Parameters
	----------
	model_class : class
		The model_class should be implement the ``fit(x, y)`` and
		``predict(x)`` methods, as used by the regression models
		present in the scikit-learn library.

	err_func : callable
		Scorer callable object with signature ``score(estimator, x, y)``.

	inverse_error_func : callable
		Inverse (or partial inverse) of err_func.

	model_params : dict, optional
		Dict containing keyword parameters to pass to model_class upon
		initialization.

	Attributes
	----------
	model_class : class
		Class used to construct the underlying model.

	err_func : callable
		Scorer function used to calculate nonconformity scores.

	inverse_err_func : callable
		Inverse function (partial) of nonconformity function.

	model_params : dict
		Parameters sent to underlying model.

	model : object
		Underlying model object.

	See also
	--------
	ProbEstClassifierNc
	"""
	def __init__(self,
	             model_class,
	             err_func,
	             inverse_err_func,
	             model_params=None):
		super(RegressorNc, self).__init__(model_class,
		                                  err_func,
		                                  model_params)

		self.inverse_err_func = inverse_err_func

	def _underlying_predict(self, x):
		return self.model.predict(x)

	def predict(self, x, nc, significance):
		prediction = self._underlying_predict(x)
		return self.inverse_err_func(prediction, nc, significance)


#!/usr/bin/env python

"""
Inductive conformal predictors.
"""

# Authors: Henrik Linusson

from __future__ import division

import numpy as np

# -----------------------------------------------------------------------------
# Base inductive conformal predictor
# -----------------------------------------------------------------------------
class BaseIcp(object):
	"""Base class for inductive conformal predictors.
	"""

	_problem_type = None

	@classmethod
	def get_problem_type(cls):
		"""Problem type of conformal predictor.

		Returns
		-------
			problem_type : string or None
				None, 'classification' or 'regression'
		"""
		return cls._problem_type

	def __init__(self, nc_function):
		self.cal_x, self.cal_y = None, None
		self.nc_function = nc_function

	def fit(self, x, y):
		"""Fit underlying nonconformity scorer.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for fitting the nonconformity scorer.

		y : numpy array of shape [n_samples]
			Outputs of examples for fitting the nonconformity scorer.

		Returns
		-------
		None
		"""
		#TODO: incremental?
		self.nc_function.fit(x, y)

	def calibrate(self, x, y, increment=False):
		"""Calibrate conformal predictor based on underlying nonconformity
		scorer.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for calibrating the conformal predictor.

		y : numpy array of shape [n_samples, n_features]
			Outputs of examples for calibrating the conformal predictor.

		increment : boolean
			If ``True``, performs an incremental recalibration of the conformal
			predictor. The supplied ``x`` and ``y`` are added to the set of
			previously existing calibration examples, and the conformal
			predictor is then calibrated on both the old and new calibration
			examples.

		Returns
		-------
		None
		"""
		# TODO: conditional
		self._calibrate_hook(x, y, increment)
		self._update_calibration_set(x, y, increment)
		self.cal_scores = self.nc_function.calc_nc(self.cal_x, self.cal_y)

	def _calibrate_hook(self, x, y, increment):
		pass

	def _update_calibration_set(self, x, y, increment):
		if increment and self.cal_x is not None and self.cal_y is not None:
			self.cal_x = np.vstack([self.cal_x, x])
			self.cal_y = np.hstack([self.cal_y, y])
		else:
			self.cal_x, self.cal_y = x, y

# -----------------------------------------------------------------------------
# Inductive conformal classifier
# -----------------------------------------------------------------------------
class IcpClassifier(BaseIcp):
	"""Inductive conformal classifier.

	Parameters
	----------
	nc_function : object
		Nonconformity scorer object used to calculate nonconformity of
		calibration examples and test patterns. Should implement ``fit(x, y)``
		and ``calc_nc(x, y)``.

	smoothing : boolean
		Decides whether to use stochastic smoothing of p-values.

	Attributes
	----------
	cal_x : numpy array of shape [n_cal_examples, n_features]
		Inputs of calibration set.

	cal_y : numpy array of shape [n_cal_examples]
		Outputs of calibration set.

	nc_function : object
		Nonconformity scorer object used to calculate nonconformity scores.

	See also
	--------
	IcpRegressor

	References
	----------
	.. [1] Papadopoulos, H., & Haralambous, H. (2011). Reliable prediction
		intervals with regression neural networks. Neural Networks, 24(8),
		842-851.

	Examples
	--------
	>>> import numpy as np
	>>> from sklearn.datasets import load_iris
	>>> from sklearn.tree import DecisionTreeClassifier
	>>> from nonconformist.icp import IcpClassifier
	>>> from nonconformist.nc import ProbEstClassifierNc, margin
	>>> iris = load_iris()
	>>> idx = np.random.permutation(iris.target.size)
	>>> train = idx[:int(idx.size / 3)]
	>>> cal = idx[int(idx.size / 3):int(2 * idx.size / 3)]
	>>> test = idx[int(2 * idx.size / 3):]
	>>> nc = ProbEstClassifierNc(DecisionTreeClassifier, margin)
	>>> icp = IcpClassifier(nc)
	>>> icp.fit(iris.data[train, :], iris.target[train])
	>>> icp.calibrate(iris.data[cal, :], iris.target[cal])
	>>> icp.predict(iris.data[test, :], significance=0.10)
	...             # doctest: +SKIP
	array([[ True, False, False],
		[False,  True, False],
		...,
		[False,  True, False],
		[False,  True, False]], dtype=bool)
	"""

	_problem_type = 'classification'

	def __init__(self, nc_function, smoothing=True):
		super(IcpClassifier, self).__init__(nc_function)
		self.classes = None
		self.last_p = None
		self.smoothing = smoothing

	def _calibrate_hook(self, x, y, increment=False):
		self._update_classes(y, increment)

	def _update_classes(self, y, increment):
		if self.classes is None or not increment:
			self.classes = np.unique(y)
		else:
			self.classes = np.unique(np.hstack([self.classes, y]))

	def predict(self, x, significance=None):
		"""Predict the output values for a set of input patterns.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		significance : float or None
			Significance level (maximum allowed error rate) of predictions.
			Should be a float between 0 and 1. If ``None``, then the p-values
			are output rather than the predictions.

		Returns
		-------
		p : numpy array of shape [n_samples, n_classes]
			If significance is ``None``, then p contains the p-values for each
			sample-class pair; if significance is a float between 0 and 1, then
			p is a boolean array denoting which labels are included in the
			prediction sets.
		"""
		n_test_objects = x.shape[0]
		p = np.zeros((n_test_objects, self.classes.size))
		for i, c in enumerate(self.classes):
			test_class = np.zeros(x.shape[0])
			test_class.fill(c)

			# TODO: maybe calculate p-values using cython or similar
			# TODO: interpolated p-values

			test_nc_scores = self.nc_function.calc_nc(x, test_class)
			n_cal = self.cal_scores.size
			for j, nc in enumerate(test_nc_scores):
				n_ge = np.sum(self.cal_scores >= nc)
				p[j, i] = n_ge / (n_cal + 1)

			if self.smoothing:
				p[:, i] += np.random.uniform(0, 1, n_test_objects) / (n_cal + 1)
			else:
				p[:, i] += 1 / (n_cal + 1)

		if significance:
			return p > significance
		else:
			return p

# -----------------------------------------------------------------------------
# Inductive conformal regressor
# -----------------------------------------------------------------------------
class IcpRegressor(BaseIcp):
	"""Inductive conformal regressor.

	Parameters
	----------
	nc_function : object
		Nonconformity scorer object used to calculate nonconformity of
		calibration examples and test patterns. Should implement ``fit(x, y)``,
		``calc_nc(x, y)`` and ``predict(x, nc_scores, significance)``.

	Attributes
	----------
	cal_x : numpy array of shape [n_cal_examples, n_features]
		Inputs of calibration set.

	cal_y : numpy array of shape [n_cal_examples]
		Outputs of calibration set.

	nc_function : object
		Nonconformity scorer object used to calculate nonconformity scores.

	See also
	--------
	IcpClassifier

	References
	----------
	.. [1] Papadopoulos, H., Proedrou, K., Vovk, V., & Gammerman, A. (2002).
		Inductive confidence machines for regression. In Machine Learning: ECML
		2002 (pp. 345-356). Springer Berlin Heidelberg.

	.. [2] Papadopoulos, H., & Haralambous, H. (2011). Reliable prediction
		intervals with regression neural networks. Neural Networks, 24(8),
		842-851.

	Examples
	--------
	>>> import numpy as np
	>>> from sklearn.datasets import load_boston
	>>> from sklearn.tree import DecisionTreeRegressor
	>>> from nonconformist.icp import IcpRegressor
	>>> from nonconformist.nc import RegressorNc, abs_error, abs_error_inv
	>>> boston = load_boston()
	>>> idx = np.random.permutation(boston.target.size)
	>>> train = idx[:int(idx.size / 3)]
	>>> cal = idx[int(idx.size / 3):int(2 * idx.size / 3)]
	>>> test = idx[int(2 * idx.size / 3):]
	>>> nc = RegressorNc(DecisionTreeRegressor, abs_error, abs_error_inv)
	>>> icp = IcpRegressor(nc)
	>>> icp.fit(boston.data[train, :], boston.target[train])
	>>> icp.calibrate(boston.data[cal, :], boston.target[cal])
	>>> icp.predict(boston.data[test, :], significance=0.10)
	...     # doctest: +SKIP
	array([[  5. ,  20.6],
		[ 15.5,  31.1],
		...,
		[ 14.2,  29.8],
		[ 11.6,  27.2]])
	"""

	_problem_type = 'regression'

	def __init__(self, nc_function):
		super(IcpRegressor, self).__init__(nc_function)

	def predict(self, x, significance):
		"""Predict the output values for a set of input patterns.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		significance : float
			Significance level (maximum allowed error rate) of predictions.
			Should be a float between 0 and 1.

		Returns
		-------
		p : numpy array of shape [n_samples, 2]
			Prediction interval (minimum and maximum boundaries) for
			the set of test patterns.
		"""
		# TODO: interpolated p-values
		return self.nc_function.predict(x, self.cal_scores, significance)


#!/usr/bin/env python

"""
Aggregated conformal predictors
"""

# Authors: Henrik Linusson

import numpy as np
from sklearn.cross_validation import KFold, StratifiedKFold
from sklearn.cross_validation import ShuffleSplit, StratifiedShuffleSplit

# -----------------------------------------------------------------------------
# Sampling strategies
# -----------------------------------------------------------------------------
class BootstrapSampler(object):
	"""Bootstrap sampler.

	See also
	--------
	CrossSampler, RandomSubSampler

	Examples
	--------
	"""
	def gen_samples(self, x, y, n_samples, problem_type):
		for i in range(n_samples):
			idx = np.array(range(y.size))
			train = np.random.choice(y.size, y.size, replace=True)
			cal_mask = np.array(np.ones(idx.size), dtype=bool)
			for j in train:
				cal_mask[j] = False
			cal = idx[cal_mask]

			yield train, cal

class CrossSampler(object):
	"""Cross-fold sampler.

	See also
	--------
	BootstrapSampler, RandomSubSampler

	Examples
	--------
	"""
	def gen_samples(self, x, y, n_samples, problem_type):
		if problem_type == 'classification':
			folds = StratifiedKFold(y, n_folds=n_samples)
		else:
			folds = KFold(y.size, n_folds=n_samples)
		for train, cal in folds:
			yield train, cal

class RandomSubSampler(object):
	"""Random subsample sampler.

	Parameters
	----------
	calibration_portion : float
		Ratio (0-1) of examples to use for calibration.

	See also
	--------
	BootstrapSampler, CrossSampler

	Examples
	--------
	"""
	def __init__(self, calibration_portion=0.3):
		self.cal_portion = calibration_portion

	def gen_samples(self, x, y, n_samples, problem_type):
		if problem_type == 'classification':
			splits = StratifiedShuffleSplit(y,
			                                n_iter=n_samples,
			                                test_size=self.cal_portion)
		else:
			splits = ShuffleSplit(y.size,
			                      n_iter=n_samples,
			                      test_size=self.cal_portion)

		for train, cal in splits:
			yield train, cal

# -----------------------------------------------------------------------------
# Conformal ensemble
# -----------------------------------------------------------------------------
class AggregatedCp(object):
	"""Aggregated conformal predictor.

	Combines multiple IcpClassifier or IcpRegressor predictors into an
	aggregated model.


	Parameters
	----------
	cp_class : class
		Class of conformal predictor to use (e.g. IcpClassifier or IcpRegressor)

	nc_class : 	Nonconformity scorer object used to calculate nonconformity of
		calibration examples and test patterns. Should implement ``fit(x, y)``
		and ``calc_nc(x, y)``. For regression problems, it should also
		implement ``predict(x, nc_scores, significance)``.

	sampler : object
		Sampler object used to generate training and calibration examples
		for the underlying conformal predictors.

	aggregation_func : callable
		Function used to aggregate the predictions of the underlying
		conformal predictors. Defaults to ``numpy.mean``.

	nc_class_params : dict, optional
		Parameters to pass to nc_class.

	n_models : int
		Number of models to aggregate.

	Attributes
	----------
	predictors : list
		List of underlying conformal predictors.

	sampler : object
		Sampler object used to generate training and calibration examples.

	agg_func : callable
		Function used to aggregate the predictions of the underlying
		conformal predictors

	References
	----------
	.. [1] Vovk, V. (2013). Cross-conformal predictors. Annals of Mathematics
		and Artificial Intelligence, 1-20.

	.. [2] Carlsson, L., Eklund, M., & Norinder, U. (2014). Aggregated
		Conformal Prediction. In Artificial Intelligence Applications and
		Innovations (pp. 231-240). Springer Berlin Heidelberg.

	Examples
	--------
	"""
	def __init__(self,
	             cp_class,
	             nc_class,
	             sampler=BootstrapSampler,
	             aggregation_func=None,
	             nc_class_params=None,
	             n_models=10):
		self.predictors = []
		self.n_models = n_models
		self.cp_class = cp_class
		self.nc_class = nc_class
		self.nc_class_params = nc_class_params if nc_class_params else {}
		self.sampler = sampler

		if aggregation_func is not None:
			self.agg_func = aggregation_func
		else:
			self.agg_func = lambda x: np.mean(x, axis=2)

	def fit(self, x, y):
		"""Fit underlying conformal predictors.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for fitting the underlying conformal predictors.

		y : numpy array of shape [n_samples]
			Outputs of examples for fitting the underlying conformal predictors.

		Returns
		-------
		None
		"""
		self.predictors = []
		idx = np.random.permutation(y.size)
		x, y = x[idx, :], y[idx]
		samples = self.sampler.gen_samples(x,
		                                   y,
		                                   self.n_models,
		                                   self.cp_class.get_problem_type())
		for train, cal in samples:
			predictor = self.cp_class(self.nc_class(**self.nc_class_params))
			predictor.fit(x[train, :], y[train])
			predictor.calibrate(x[cal, :], y[cal])
			self.predictors.append(predictor)

	def predict(self, x, significance=None):
		"""Predict the output values for a set of input patterns.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		significance : float or None
			Significance level (maximum allowed error rate) of predictions.
			Should be a float between 0 and 1. If ``None``, then the p-values
			are output rather than the predictions. Note: ``significance=None``
			is applicable to classification problems only.

		Returns
		-------
		p : numpy array of shape [n_samples, n_classes] or [n_samples, 2]
			For classification problems: If significance is ``None``, then p
			contains the p-values for each sample-class pair; if significance
			is a float between 0 and 1, then p is a boolean array denoting
			which labels are included in the prediction sets.

			For regression problems: Prediction interval (minimum and maximum
			boundaries) for the set of test patterns.
		"""
		is_regression = self.cp_class.get_problem_type() == 'regression'

		f = lambda p, x: p.predict(x, significance if is_regression else None)
		predictions = np.dstack([f(p, x) for p in self.predictors])
		predictions = self.agg_func(predictions)

		if significance and not is_regression:
			return predictions >= significance
		else:
			return predictions


#!/usr/bin/env python

"""
Evaluation of conformal predictors.
"""

# Authors: Henrik Linusson

# TODO: should possibly (re)implement a cross-validator suited for CP

from __future__ import division

import numpy as np

from sklearn.cross_validation import StratifiedShuffleSplit
from sklearn.cross_validation import train_test_split

class IcpClassCrossValHelper(object):
	"""Helper class for running sklearn's ``cross_val_score`` evaluation
	method on IcpClassifiers.

	See also
	--------
	IcpRegCrossValHelper

	Examples
	--------
	>>> from sklearn.datasets import load_iris
	>>> from sklearn.ensemble import RandomForestClassifier
	>>> from sklearn.cross_validation import cross_val_score
	>>> from nonconformist.icp import IcpClassifier
	>>> from nonconformist.nc import ProbEstClassifierNc, margin
	>>> from nonconformist.evaluation import IcpClassCrossValHelper
	>>> from nonconformist.evaluation import class_mean_errors
	>>> data = load_iris()
	>>> nc = ProbEstClassifierNc(RandomForestClassifier, margin)
	>>> icp = IcpClassifier(nc)
	>>> icp_cv = IcpClassCrossValHelper(icp, significance=0.1)
	>>> cross_val_score(icp_cv,
	...                 data.data,
	...                 data.target,
	...                 scoring=class_mean_errors,
	...                 cv=10)
	...     # doctest: +SKIP
	array([ 0.        ,  0.        ,  0.        ,  0.06666667,  0.33333333,
	0.2       ,  0.        ,  0.13333333,  0.53333333,  0.06666667])
	"""
	def __init__(self, icp, calibration_portion=0.25, significance=0.1):
		self.icp = icp
		self.cal_port = calibration_portion
		self.significance = significance

	def fit(self, x, y):
		split = StratifiedShuffleSplit(y, n_iter=1, test_size=self.cal_port)
		for train, cal in split:
			self.icp.fit(x[train, :], y[train])
			self.icp.calibrate(x[cal, :], y[cal])

	def predict(self, x, significance=True):
		if significance:
			return self.icp.predict(x, significance=self.significance)
		else:
			return self.icp.predict(x, significance=None)

	def get_params(self, deep=False):
		return {'icp': self.icp,
		        'calibration_portion': self.cal_port,
		        'significance': self.significance}

class IcpRegCrossValHelper(object):
	"""Helper class for running sklearn's ``cross_val_score`` evaluation
	method on IcpRegressors.

	See also
	--------
	IcpClassCrossValHelper

	Examples
	--------
	>>> from sklearn.datasets import load_boston
	>>> from sklearn.ensemble import RandomForestRegressor
	>>> from sklearn.cross_validation import cross_val_score
	>>> from nonconformist.icp import IcpRegressor
	>>> from nonconformist.nc import RegressorNc, abs_error, abs_error_inv
	>>> from nonconformist.evaluation import IcpRegCrossValHelper
	>>> from nonconformist.evaluation import reg_mean_errors
	>>> data = load_boston()
	>>> nc = RegressorNc(RandomForestRegressor, abs_error, abs_error_inv)
	>>> icp = IcpRegressor(nc)
	>>> icp_cv = IcpRegCrossValHelper(icp, significance=0.1)
	>>> cross_val_score(icp_cv,
	...                 data.data,
	...                 data.target,
	...                 scoring=reg_mean_errors,
	...                 cv=10)
	...     # doctest: +SKIP
	array([ 0.17647059,  0.03921569,  0.        ,  0.23529412,  0.11764706,
	0.17647059,  0.2       ,  0.36      ,  0.2       ,  0.16      ])
	"""
	def __init__(self, icp, calibration_portion=0.25, significance=0.1):
		self.icp = icp
		self.cal_port = calibration_portion
		self.significance = significance

	def fit(self, x, y):
		split = train_test_split(x, y, test_size=self.cal_port)
		x_tr, x_cal, y_tr, y_cal = split[0], split[1], split[2], split[3]
		self.icp.fit(x_tr, y_tr)
		self.icp.calibrate(x_cal, y_cal)

	def predict(self, x, significance=True):
		if significance:
			return self.icp.predict(x, significance=self.significance)
		else:
			return self.icp.predict(x, significance=None)

	def get_params(self, deep=False):
		return {'icp': self.icp,
		        'calibration_portion': self.cal_port,
		        'significance': self.significance}

# -----------------------------------------------------------------------------
# Validity measures
# -----------------------------------------------------------------------------
def reg_n_correct(model, x, y, significance=None):
	"""Calculates the number of correct predictions made by a conformal
	regression model.

	Parameters
	----------
	model : object
		Conformal regressor object.

	x : numpy array of shape [n_samples, n_features]
		Inputs of test objects.

	y : numpy array of shape [n_samples]
		Outputs of test objects.
	"""
	if significance:
		prediction = model.predict(x, significance)
	else:
		prediction = model.predict(x)
	low = y >= prediction[:,0]
	high = y <= prediction[:,1]
	correct = low * high

	return y[correct].size

def reg_mean_errors(model, x, y, significance=None):
	"""Calculates the average error rate of a conformal regression model.

	Parameters
	----------
	model : object
		Conformal regressor object.

	x : numpy array of shape [n_samples, n_features]
		Inputs of test objects.

	y : numpy array of shape [n_samples]
		Outputs of test objects.
	"""
	return 1 - reg_n_correct(model, x, y, significance) / y.size

def class_n_correct(model, x, y, significance=None):
	"""Calculates the number of correct predictions made by a conformal
	classification model.

	Parameters
	----------
	model : object
		Conformal classifier object.

	x : numpy array of shape [n_samples, n_features]
		Inputs of test objects.

	y : numpy array of shape [n_samples]
		Outputs of test objects.
	"""
	if significance:
		prediction = model.predict(x, significance)
	else:
		prediction = model.predict(x)
	correct = np.zeros((y.size,), dtype=bool)
	for i, y_ in enumerate(y):
		correct[i] = prediction[i, y_]
	return np.sum(correct)

def class_mean_errors(model, x, y, significance=None):
	"""Calculates the average error rate of a conformal classification model.

	Parameters
	----------
	model : object
		Conformal classifier object.

	x : numpy array of shape [n_samples, n_features]
		Inputs of test objects.

	y : numpy array of shape [n_samples]
		Outputs of test objects.
	"""
	return 1 - (class_n_correct(model, x, y, significance) / y.size)

# -----------------------------------------------------------------------------
# Efficiency measures
# -----------------------------------------------------------------------------
def reg_mean_size(model, x, y, significance=None):
	"""Calculates the average prediction interval size of a conformal
	regression model.

	Parameters
	----------
	model : object
		Conformal regressor object.

	x : numpy array of shape [n_samples, n_features]
		Inputs of test objects.

	y : numpy array of shape [n_samples]
		Outputs of test objects.
	"""
	if significance:
		prediction = model.predict(x)
	else:
		prediction = model.predict(x, significance)
	interval_size = 0
	for j in range(y.size):
		interval_size += np.abs(prediction[j, 1] - prediction[j, 0])
	return interval_size / y.size

def class_avg_c(model, x, y, significance=None):
	"""Calculates the average number of classes per prediction of a conformal
	classification model.

	Parameters
	----------
	model : object
		Conformal classifier object.

	x : numpy array of shape [n_samples, n_features]
		Inputs of test objects.

	y : numpy array of shape [n_samples]
		Outputs of test objects.
	"""
	if significance:
		prediction = model.predict(x, significance)
	else:
		prediction = model.predict(x)
	return np.sum(prediction) / prediction.shape[0]

def class_mean_p_val(model, x, y):
	"""Calculates the mean of the p-values output by a conformal classification
	model.

	Parameters
	----------
	model : object
		Conformal classifier object.

	x : numpy array of shape [n_samples, n_features]
		Inputs of test objects.

	y : numpy array of shape [n_samples]
		Outputs of test objects.
	"""
	prediction = model.predict(x, significance=None)
	return np.mean(prediction)

def class_one_c(model, x, y, significance=None):
	"""Calculates the rate of singleton predictions (prediction sets containing
	only a single class label) of a conformal classification model.

	Parameters
	----------
	model : object
		Conformal classifier object.

	x : numpy array of shape [n_samples, n_features]
		Inputs of test objects.

	y : numpy array of shape [n_samples]
		Outputs of test objects.
	"""
	if significance:
		prediction = model.predict(x, significance)
	else:
		prediction = model.predict(x)
	return np.sum(1 for _ in filter(lambda x: np.sum(x) == 1, prediction))